package pwc.taxtech.atms.vat.service.impl;

import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Service;
import pwc.taxtech.atms.common.datasource.ShardingContextHolder;
import pwc.taxtech.atms.constant.enums.EnumServiceType;
import pwc.taxtech.atms.dto.OperationResultDto;
import pwc.taxtech.atms.dto.vatdto.CellTemplateReferenceDto;
import pwc.taxtech.atms.dto.vatdto.ReportDto;
import pwc.taxtech.atms.entitiy.*;
import pwc.taxtech.atms.exception.NotSupportedException;
import pwc.taxtech.atms.vat.entity.PeriodCellTemplate;
import pwc.taxtech.atms.vat.entity.PeriodCellTemplateConfig;
import pwc.taxtech.atms.vat.entity.PeriodTemplate;
import pwc.taxtech.atms.vat.entity.PeriodTemplateExample;
import pwc.taxtech.atms.vat.service.ReportService;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

@Service
public class ReportServiceImpl extends VatAbstractService implements ReportService {
    @Override
    public OperationResultDto<List<ReportDto>> getReportTemplate(String projectID, EnumServiceType serviceType, Integer periodParam) {
        int period = periodParam != null ? periodParam : 0;
        OperationResultDto<List<ReportDto>> operationResult = new OperationResultDto<>();
        Project project = projectMapper.selectByPrimaryKey(projectID);
        String organizationID = project == null ? null : project.getOrganizationID();
        if (organizationID == null) {
            operationResult.setResult(false);
            operationResult.setResultMsg("NoOrganization");
            return operationResult;
        }

        Long templateGroupID;

        /*
        对于VAT，从Rule Engine中获取TemplateGroupID
        对于其他Service Type，直接从ProjectServiceType中获取
         */
        if (serviceType.equals(EnumServiceType.VAT)) {
            TaxPayerReportRuleExample example = new TaxPayerReportRuleExample();
            example.createCriteria().andOrganizationIdLike(organizationID);
            List<TaxPayerReportRule> taxPayerReportRules = taxPayerReportRuleMapper.selectByExample(example);
            Optional<TaxPayerReportRule> taxPayerReportRule = taxPayerReportRules.stream().findFirst();
            if (!taxPayerReportRule.isPresent()) {
                Optional<TaxPayerReportRule> defaultRule = taxPayerReportRules.stream().filter(x -> x.getIsDefault()).findFirst();
                if (!defaultRule.isPresent()) {
                    operationResult.setResultMsg("NoDefaultRule");
                    return operationResult;
                }

                templateGroupID = defaultRule.get().getTemplateGroupId();
            } else {
                templateGroupID = taxPayerReportRule.get().getTemplateGroupId();
            }

        } else {
            String serviceTypeStr = serviceType.getCode().toString();
            ProjectServiceTypeExample projectServiceTypeExample = new ProjectServiceTypeExample();
            projectServiceTypeExample.createCriteria().andServiceTypeIDEqualTo(serviceTypeStr).andProjectIDEqualTo(projectID);
            Optional<Long> tempVlaue = projectServiceTypeMapper.selectByExample(projectServiceTypeExample).stream().map(ProjectServiceType::getTemplateGroupID).findFirst();
            if(tempVlaue.isPresent()){
                templateGroupID = tempVlaue.get();
            }
            else{
                templateGroupID = 0L;
            }
        }

        String dbName = ShardingContextHolder.getDataSourceKey();
        Map<String, Object> map = new HashMap<>();
        map.put("dbName", dbName);
        map.put("period", period);
        map.put("projectID", projectID);
        map.put("templateGroupID", templateGroupID);

        List<ReportDto> reportDtos = templateMapper.getTemplateLeftJoinReport(map);
        if (reportDtos.isEmpty()) {
            operationResult.setResultMsg("No Template");
            return operationResult;
        }

        operationResult.setResult(true);
        operationResult.setData(reportDtos);
        return operationResult;
    }

    @Override
    public OperationResultDto updateConfig(String projectId, Integer period, Boolean ifDeleteManualDataSource, String generator) {
        OperationResultDto result = new OperationResultDto();

        if (period == null) {
            result.setResultMsg("peirod is null");
            return result;
        }

        Project project = projectMapper.selectByPrimaryKey(projectId);

        if (project == null) {
            result.setResultMsg("NoProject");
            return result;
        }

        List<Long> exceptTemplateIDs = templateMapper.getIDsForExceptTemplate();

        String strExceptTemplateIDs = StringUtils.join(exceptTemplateIDs, ",");

        //todo: according to projectId get the templateGroupId
        Long templateGroupID = projectMapper.getTemplateGroupIdByProject(projectId, EnumServiceType.VAT.getCode());
        if (templateGroupID != null && templateGroupID != 0) {
            // 先进行数据清理，包括period开头的所有报表配置表 条件Period
            dataSourceMapper.clearFormulaBlockWithPeriod(period);
            dataSourceMapper.clearTaxRuleSettingWithPeriod(period);
            dataSourceMapper.clearCellTemplateWithPeriod(period, strExceptTemplateIDs);
            dataSourceMapper.clearCellTemplateConfigWithPeriod(period, strExceptTemplateIDs);
            dataSourceMapper.clearTemplateWithPeriod(period, strExceptTemplateIDs);
            dataSourceMapper.clearTaxPayerReportWithPeriod(period);
            dataSourceMapper.clearDataSourceWithPeriod(period);
            dataSourceMapper.clearDataSourceDetailWithPeriod(period);
            dataSourceMapper.clearCellDataWithPeriod(strExceptTemplateIDs);

            // 根据templategroupid 把 template 插入到 periodTemplate
            TemplateExample example = new TemplateExample();
            example.createCriteria().andTemplateGroupIdEqualTo(templateGroupID);
            List<Template> templates = templateMapper.selectByExample(example);
            List<PeriodTemplate> periodTemplateList = new ArrayList<>();
            for (Template template : templates) {
                PeriodTemplate periodTemplate = new PeriodTemplate();
                periodTemplate.setId(distributedIDService.nextId());
                periodTemplate.setPeriod(period);
                periodTemplate.setName(template.getName());
                periodTemplate.setCode(template.getCode());
                periodTemplate.setPath(template.getPath());
                periodTemplate.setReportType(template.getReportType());
                periodTemplate.setTemplateGroupId(template.getTemplateGroupId());
                periodTemplate.setOrderIndex(template.getOrderIndex());
                periodTemplate.setCreateTime(template.getCreateTime());
                periodTemplate.setUpdateTime(template.getUpdateTime());
                periodTemplate.setIsSystemType(template.getIsSystemType());
                periodTemplate.setIsActiveAssociation(template.getIsActiveAssociation());
                periodTemplate.setParentId(template.getParentId());
                periodTemplate.setTemplateId(template.getId());
                periodTemplate.setCreateBy(template.getCreateBy());
                periodTemplate.setUpdateBy(template.getUpdateBy());
                periodTemplateList.add(periodTemplate);
            }

            List<Long> templateIDList = periodTemplateList.stream().map(PeriodTemplate::getTemplateId).collect(Collectors.toList());
            CellTemplateExample cellTemplateExample = new CellTemplateExample();
            cellTemplateExample.createCriteria().andReportTemplateIdIn(templateIDList);
            List<CellTemplate> cellTemplateList = cellTemplateMapper.selectByExample(cellTemplateExample);
            List<PeriodCellTemplate> periodCellTemplateList = new ArrayList<>();
            for (CellTemplate cellTemplate : cellTemplateList) {
                PeriodCellTemplate periodCellTemplate = new PeriodCellTemplate();
                periodCellTemplate.setId(distributedIDService.nextId());
                periodCellTemplate.setPeriod(period);
                periodCellTemplate.setReportTemplateId(cellTemplate.getReportTemplateId());
                periodCellTemplate.setRowIndex(cellTemplate.getRowIndex());
                periodCellTemplate.setRowName(cellTemplate.getRowName());
                periodCellTemplate.setColumnIndex(cellTemplate.getColumnIndex());
                periodCellTemplate.setColumnName(cellTemplate.getColumnName());
                periodCellTemplate.setComment(cellTemplate.getComment());
                periodCellTemplate.setCreateTime(cellTemplate.getCreateTime());
                periodCellTemplate.setUpdateTime(cellTemplate.getUpdateTime());
                periodCellTemplate.setCellTemplateId(cellTemplate.getId());
                periodCellTemplate.setDataType(cellTemplate.getDataType());
                periodCellTemplate.setIsReadOnly(cellTemplate.getIsReadOnly());
                periodCellTemplate.setCopyFromId(cellTemplate.getCopyFromId());
                periodCellTemplate.setCreateBy(cellTemplate.getCreateBy());
                periodCellTemplate.setUpdateBy(cellTemplate.getUpdateBy());
                periodCellTemplateList.add(periodCellTemplate);
            }

            List<Long> cellTemplateIDList = periodCellTemplateList.stream().map(PeriodCellTemplate::getCellTemplateId).collect(Collectors.toList());
            CellTemplateConfigExample cellTemplateConfigExample = new CellTemplateConfigExample();
            cellTemplateConfigExample.createCriteria().andReportTemplateIdIn(templateIDList).andCellTemplateIdIn(cellTemplateIDList);
            List<CellTemplateConfig> cellTemplateConfigList = cellTemplateConfigMapper.selectByExample(cellTemplateConfigExample);
            List<PeriodCellTemplateConfig> periodCellTemplateConfigList = new ArrayList<>();
            for (CellTemplateConfig cellTemplateConfig : cellTemplateConfigList) {
                PeriodCellTemplateConfig periodCellTemplateConfig = new PeriodCellTemplateConfig();
                periodCellTemplateConfig.setId(distributedIDService.nextId());
                periodCellTemplateConfig.setPeriod(period);
                periodCellTemplateConfig.setCellTemplateId(cellTemplateConfig.getCellTemplateId());
                periodCellTemplateConfig.setReportTemplateId(cellTemplateConfig.getReportTemplateId());
                periodCellTemplateConfig.setDataSourceType(cellTemplateConfig.getDataSourceType());
                periodCellTemplateConfig.setFormula(cellTemplateConfig.getFormula());
                periodCellTemplateConfig.setParsedFormula(StringUtils.EMPTY);
                periodCellTemplateConfig.setFormulaDescription(cellTemplateConfig.getFormulaDescription());
                periodCellTemplateConfig.setAccountCodes(cellTemplateConfig.getAccountCodes());
                periodCellTemplateConfig.setInvoiceType(cellTemplateConfig.getInvoiceType());
                periodCellTemplateConfig.setTaxRate(cellTemplateConfig.getTaxRate());
                periodCellTemplateConfig.setInvoiceAmountType(cellTemplateConfig.getInvoiceAmountType());
                periodCellTemplateConfig.setModelIds(cellTemplateConfig.getModelIds());
                periodCellTemplateConfig.setCreateBy(cellTemplateConfig.getCreateBy());
                periodCellTemplateConfig.setCreateTime(cellTemplateConfig.getCreateTime());
                periodCellTemplateConfig.setUpdateBy(cellTemplateConfig.getUpdateBy());
                periodCellTemplateConfig.setUpdateTime(cellTemplateConfig.getUpdateTime());
                periodCellTemplateConfig.setInvoiceCategory(cellTemplateConfig.getInvoiceCategory());
                periodCellTemplateConfig.setFormulaDataSource(cellTemplateConfig.getFormulaDataSource());
                periodCellTemplateConfig.setValidation(cellTemplateConfig.getValidation());
                periodCellTemplateConfig.setParsedValidation(StringUtils.EMPTY);
                periodCellTemplateConfig.setValidationDescription(cellTemplateConfig.getValidationDescription());
                periodCellTemplateConfig.setVoucherKeyword(cellTemplateConfig.getVoucherKeyword());
                periodCellTemplateConfig.setCellTemplateConfigId(cellTemplateConfig.getId());
                periodCellTemplateConfigList.add(periodCellTemplateConfig);
            }

            if (periodCellTemplateConfigList != null) {
                periodCellTemplateConfigList.forEach(a -> {
                    if (StringUtils.isNotBlank(a.getAccountCodes())) {
                        String[] acctCodes = ",".split(a.getAccountCodes());
                        EnterpriseAccountExample enterpriseAccountExample = new EnterpriseAccountExample();
                        enterpriseAccountExample.createCriteria().andStdCodeIn(Arrays.asList(acctCodes));
                        List<EnterpriseAccount> enterpriseAccounts = enterpriseAccountMapper.selectByExample(enterpriseAccountExample);
                        if (!enterpriseAccounts.isEmpty()) {
                            List<String> acctCodeList = enterpriseAccounts.stream().map(EnterpriseAccount::getCode).collect(Collectors.toList());
                            StringBuilder codes = new StringBuilder();
                            for (String code : acctCodeList) {
                                codes.append(code).append(",");
                            }
                            a.setAccountCodes(StringUtils.removeEnd(codes.toString(), ","));
                        }
                    }
                });
            }

            //todo: save config data to DB
            periodTemplateMapper.batchInsert(periodTemplateList);
            periodCellTemplateMapper.batchInsert(periodCellTemplateList);
            periodCellTemplateConfigMapper.batchInsert(periodCellTemplateConfigList);
        } else {
            result.setResult(true);
            result.setResultMsg("there is no templateGroup");
            return result;
        }
        result.setResult(true);
        return result;
    }

    @Override
    public OperationResultDto generateData(String projectId, EnumServiceType serviceType, Boolean ifDeleteManualDataSource, Integer periodParam, Integer reportType, String generator) {
        OperationResultDto operationResultDto = new OperationResultDto();

        if (serviceType.equals(EnumServiceType.VAT) && (periodParam == null && periodParam <= 0)) {
            operationResultDto.setResultMsg("PeriodRequiredForVAT");
            return operationResultDto;
        }

        if (serviceType != EnumServiceType.VAT) {
            // 暂不支持非CIT/VAT service批量后端生成报表
            throw new NotSupportedException();
        }

        int period = periodParam != null ? periodParam : 0;
        String serviceTypeStr = serviceType.getCode().toString();

        ProjectServiceTypeExample projectServiceTypeExample = new ProjectServiceTypeExample();
        projectServiceTypeExample.createCriteria().andServiceTypeIDEqualTo(serviceTypeStr).andProjectIDEqualTo(projectId);
        Optional<Long> templateGroupID = projectServiceTypeMapper.selectByExample(projectServiceTypeExample).stream().map(ProjectServiceType::getTemplateGroupID).findFirst();
        if (templateGroupID == null) {
            operationResultDto.setResultMsg("TemplateGroupNotExist");
            return operationResultDto;
        }

        List<Long> templateIds;
        if (reportType != null) {
            PeriodTemplateExample periodTemplateExample = new PeriodTemplateExample();
            periodTemplateExample.createCriteria().andTemplateGroupIdEqualTo(templateGroupID.get()).andPeriodEqualTo(period).andIsActiveAssociationEqualTo(true).andReportTypeEqualTo(reportType);
            templateIds = periodTemplateMapper.selectByExample(periodTemplateExample).stream().map(PeriodTemplate::getId).collect(Collectors.toList());
        } else {
            PeriodTemplateExample periodTemplateExample = new PeriodTemplateExample();
            periodTemplateExample.createCriteria().andTemplateGroupIdEqualTo(templateGroupID.get()).andPeriodEqualTo(period).andIsActiveAssociationEqualTo(true);
            templateIds = periodTemplateMapper.selectByExample(periodTemplateExample).stream().map(PeriodTemplate::getId).collect(Collectors.toList());
        }


        String rslt = reportGenerator.generateData(projectId, templateIds, ifDeleteManualDataSource, null, periodParam, generator);
        if (StringUtils.isBlank(rslt)) {
            operationResultDto.setResultMsg("ReportGenerateFailed!");
            return operationResultDto;
        }
        operationResultDto.setResult(true);
        return operationResultDto;
    }

    @Override
    public List<CellTemplateReferenceDto> getTemplateReferences(int period) {
        return new ArrayList<>();
    }
}
